dyn.load(paste0("dtr", .Platform$dynlib.ext))


# preprocessing ----------

library(data.table)

# subjects and treatment randomization

subjects <- fread("id_status_ad.csv", sep = ",")
subjects[, table(PATHWAY_ID)]
subjects[, table(DRUG_NAME, DRUG_STATUS)]

subjects[, assignment1 := ifelse(DRUG_STATUS == "PLACEBO", "Placebo", DRUG_NAME)]
subjects[, assignment2 := ifelse(DRUG_STATUS != "PLACEBO", "Default", DRUG_NAME)]

neg2na <- function(x)
{
  x[x < -3] <- NA
  x
}

# demographics

df <- fread("df_final_ad.csv", sep = ",")
length(unique(c(df[, STEPID])))

subjects <- merge(subjects[, list(STEPID, 
                                  DRUG_NAME, 
                                  DRUG_STATUS,
                                  assignment1,
                                  assignment2)], 
                  df[, list(STEPID, 
                            AGE = neg2na(AGE), 
                            GENDER = neg2na(GENDER),
                            RACE = neg2na(RACE),
                            # MARSTAT = neg2na(MARSTAT),
                            # LIVEALON = neg2na(LIVEALON), 
                            EDUCATE = neg2na(EDUCATE), 
                            EMPLOY = neg2na(EMPLOY)
                            )],
                  by = "STEPID", all.x = TRUE)

# entry

ad <- fread("ad_final_ad.csv", sep = ",")
length(unique(c(ad[, STEPID])))
ad <- ad[, lapply(.SD, max), by = STEPID]

subjects <- merge(subjects,
                  ad[, list(STEPID, 
                            BTYPE = neg2na(BTYPE),
                            PRONSET = neg2na(PRONSET))],
                  by = "STEPID", all.x = TRUE)

# exit

adx <- fread("adx_final_ad.csv", sep = ",")
length(unique(c(adx[, STEPID])))

subjects <- merge(subjects,
                  adx[, list(STEPID, EXITRZ, DEPREXIT)],
                  by = "STEPID", all.x = TRUE)

sum(complete.cases(subjects))

# clinical monitoring

cmf <- fread("cmf_final_ad.csv", sep = ",")
length(unique(c(cmf[, STEPID])))

cmf <- merge(cmf, subjects[, list(STEPID, EXITRZ)], by = "STEPID", all.x = TRUE)
cmf <- cmf[DAYSRZ <= EXITRZ, ]

# SUMD and SUMM

transform_low <- function(x) {x[x > 2 | x < -2] <- NA; ifelse(x < 0, -x, 0)}
transform_high <- function(x) {x[x > 2 | x < -2] <- NA; ifelse(x > 0, x, 0)}
transform_nonzero <- function(x) {x[x > 2 | x < -2] <- NA; abs(x)}

first <- function(x)
{
  x <- neg2na(x)
  value <- as(NA, class(x))
  exist <- !is.na(x)
  if (any(exist)) value <- x[which(exist)[1]]
  value
}

measure <- function(dt, day, suffix = "")
{
  i <- dt[, sort.list(abs(DAYSRZ - day))]
  dt2 <- dt[i, lapply(.SD, first)]
  dt2 <- dt2[, list(DAY = DAYSRZ, 
                    SUMD = transform_low(DEPINTER) + transform_low(DEPSE) + 
                      transform_low(DEPENERG) + transform_low(DEPCONCN) + 
                      transform_high(DEPGUILT) + transform_high(DEPDIST) + 
                      transform_high(DEPPMR) + transform_high(DEPPMA) + 
                      transform_high(DEPSI) + 
                      transform_nonzero(DEPSLEEP) + transform_nonzero(DEPAPPET), 
                    SUMM = transform_low(ELVSLEEP) + 
                      transform_high(ELVSELFE) + transform_high(ELVTALK) + 
                      transform_high(ELVFOI) + transform_high(ELVDISTR) + 
                      transform_high(ELVGDACT) + transform_high(ELVPMA) + 
                      transform_high(ELVHRB), 
                    STA = CLINSTAT, 
                    CGI = CGI,
                    AE_TRE = TREMORSV, 
                    AE_DRY = DRYMTHSV, 
                    AE_SED = SEDATNSV, 
                    AE_CON = CONSTISV, 
                    AE_DIA = DIARRHSV, 
                    AE_HEA = HEADACSV, 
                    AE_MEM = MEMSV, 
                    AE_SEX = SEXDYSSV, 
                    AE_APP = INCAPPSV)]
  names(dt2) <- paste0(names(dt2), suffix)
  dt2
}

baseline <- cmf[DAYSRZ <= 14, 
                measure(.SD, day = 0, suffix = "0"),
                by = STEPID]
baseline <- baseline[, list(STEPID, SUMD0, SUMM0, STA0, CGI0)]

outcome1 <- cmf[DAYSRZ >= 28 & DAYSRZ < 56, 
                measure(.SD, day = 42, suffix = "1"),
                by = STEPID]
outcome1 <- outcome1[, list(STEPID, SUMD1, SUMM1, STA1, CGI1,
                            AE_TRE1, AE_DRY1, AE_SED1, AE_CON1, AE_DIA1,
                            AE_HEA1, AE_MEM1, AE_SEX1, AE_APP1)]

outcome2 <- cmf[DAYSRZ >= 28 & DAYSRZ < 98, 
                measure(.SD, day = 84, suffix = "2"),
                by = STEPID]
outcome2 <- outcome2[, list(STEPID, SUMD2)]

subjects <- merge(subjects, baseline, by = "STEPID", all.x = TRUE)
subjects <- merge(subjects, outcome1, by = "STEPID", all.x = TRUE)
subjects <- merge(subjects, outcome2, by = "STEPID", all.x = TRUE)

subjects[, response1 := ifelse((STA1 == 7) | ((STA1 == 5) & CGI0 - CGI1 >= 1), 1,
                               ifelse(STA1 %in% c(2, 3, 4, 6), -1, 0))]
subjects[, eligible2 := ifelse(EXITRZ > 42, 1, 0)]

sum(complete.cases(subjects))

subjects[, treatment1 := assignment1]
subjects[, treatment2 := ifelse(assignment1 == "Placebo" & response1 == 0 & eligible2 == 1, 
                                assignment2, "Default")]

subjects <- subjects[complete.cases(subjects), ]


# age: integer
# gender: male(1 -> 1), female(2 -> 0), transgender(3 -> 0)
# race: white (1 -> 1), other (2, 3, 4, 5, 6 -> 0)
# marital status: married (2, 3 -> (1, 0)), never married (1 -> (0, 0)), other (4, 5, 6 -> (0, 1))
# living alone: yes (1 -> 1), no (0 -> 0)
# education: high school or below (2, 3, 4 -> 1), some college (5, 6 -> 2), bachalor and up (7, 8 -> 3)
# work: full time (1 -> 1), part time (2, 3 -> 0.5), no work (4, ..., 11 -> 0)
# bipolar type: 1, 2
# PRIOR_TO_ONSET_OF_EPISODE: Remission longer than 8 weeks (1 -> (0, 0)), 
#   Manic/hypomanic (2 -> (1, 0)), Mixed/cycling (3 -> (0, 1))
# AE_xxx: adverse event, binary

x1 <- subjects[, list(
  age = AGE,
  gender = ifelse(GENDER == 1, 1, 0),
  race = ifelse(RACE == 1, 1, 0),
  # marital1 = ifelse(MARSTAT == 1, 1, 0),
  # marital2 = ifelse(MARSTAT == 2 | MARSTAT == 3, 1, 0),
  # marital3 = ifelse(MARSTAT == 4 | MARSTAT == 5 | MARSTAT == 6, 1, 0),
  # alone = LIVEALON,
  education = ifelse(EDUCATE <= 4, 1, ifelse(EDUCATE <= 6, 2, 3)),
  work = ifelse(EMPLOY == 1, 1, ifelse(EMPLOY <= 3, 0.5, 0)),
  bipolar = BTYPE,
  prior1 = ifelse(PRONSET == 1, 1, 0),
  prior2 = ifelse(PRONSET == 2, 1, 0),
  prior3 = ifelse(PRONSET == 3, 1, 0),
  sumd0 = SUMD0,
  summ0 = SUMM0
  )]

x2 <- subjects[, list(
  ae1 = ifelse(AE_TRE1 >= 1, 1, 0),
  ae2 = ifelse(AE_DRY1 >= 1, 1, 0),
  ae3 = ifelse(AE_SED1 >= 1, 1, 0),
  ae4 = ifelse(AE_CON1 >= 1, 1, 0),
  ae5 = ifelse(AE_DIA1 >= 1, 1, 0),
  ae6 = ifelse(AE_HEA1 >= 1, 1, 0),
  ae7 = ifelse(AE_MEM1 >= 1, 1, 0),
  ae8 = ifelse(AE_SEX1 >= 1, 1, 0),
  ae9 = ifelse(AE_APP1 >= 1, 1, 0),
  sumd1 = SUMD1,
  summ1 = SUMM1
  )]

x1 <- as.data.frame(x1)
x2 <- as.data.frame(x2)

a1 <- subjects[, treatment1]
a2 <- subjects[, treatment2]

y1 <- subjects[, 0 * SUMD1]
y2 <- subjects[, SUMD2]       # lower is better


rm(list = setdiff(ls(), c("x1", "x2", "a1", "a2", "y1", "y2")))

y <- as.matrix(cbind(-y1, -y2))
a <- as.matrix(cbind(a1, a2))
x <- as.matrix(cbind(x1, x2))
stage.x <- rep(c(1, 2), c(ncol(x1), ncol(x2)))
variables <- c(names(x1), names(x2))


estimate.regime <- function(y, a, x, stage.x, seed = 1988)
{
  stopifnot(nrow(y) == nrow(a) && nrow(y) == nrow(x))
  n <- nrow(y)
  stopifnot(ncol(y) == ncol(a))
  n.stage <- ncol(y)
  dtr <- vector("list", n.stage)

  set.seed(seed)
  num.folds <- 5L
  fold <- rep_len(0L : (num.folds - 1L), n)

  future.y <- double(n)

  for (i.stage in n.stage : 1L) {
    current.x <- x[, which(stage.x <= i.stage), drop = FALSE]
    current.a <- a[, i.stage]
    current.y <- y[, i.stage] + future.y
    
    eligible <- (current.a != "Default")
    current.x <- current.x[eligible, , drop = FALSE]
    current.a <- current.a[eligible]
    current.y <- current.y[eligible]
    n <- length(current.y)
    
    if (ncol(current.x) < 2) {
      current.x <- cbind(current.x, 0)
    }

    options <- sort.int(unique(current.a))
    group <- as.integer(factor(current.a, options)) - 1L

    var.x <- colMeans(current.x ^ 2) - colMeans(current.x) ^ 2
    var.x[var.x < 1e-8] <- Inf
    scaling <- 1 / var.x
    model <- .Call("R_kernel_train", current.x, current.y,
      group, length(options), scaling)
    outcomes <- .Call("R_kernel_predict", model, current.x)
    regrets <- .Call("R_get_regrets_from_outcomes", outcomes)

    # tuning
    max.length <- 10L
    
    simple.regret <- min(colMeans(regrets))
    if (simple.regret > 1e-8) {
      zeta.grid <- simple.regret * c(2, 0.75, 0.3, 0.12, 0.05)
      eta.grid <- simple.regret * n * c(0.3, 0.1, 0.03)
      zeta.choices <- rep(zeta.grid, times = 3L)
      eta.choices <- rep(eta.grid, each = 5L)
      current.fold <- fold[sample.int(n)]
    
      cv.regret <- .Call("R_cv_tune_rule", current.x, regrets,
        zeta.choices, eta.choices, max.length,
        current.fold, num.folds)
      min.cv.regret <- min(cv.regret)
    
      if (min.cv.regret < simple.regret - 1e-8) {
        index <- which(cv.regret - min.cv.regret - 1e-8 <= 0)[1L]
        zeta.selected <- zeta.choices[index]
        eta.selected <- eta.choices[index]
      } else {
        zeta.selected <- simple.regret * n
        eta.selected <- simple.regret * n
        max.length <- 1L
      }
    } else {
      zeta.selected <- simple.regret * n
      eta.selected <- simple.regret * n
      max.length <- 1L
    }
    
    action <- integer(n)
    dtr[[i.stage]] <- .Call("R_find_rule", current.x, regrets,
      zeta.selected, eta.selected, max.length, action)
    attr(dtr[[i.stage]], "options") <- options

    future.y[eligible] <- outcomes[cbind(1L : n, action + 1L)]
    future.y[!eligible] <- future.y[!eligible] + y[!eligible, i.stage]
  }
  
  dtr
}


apply.regime <- function(current.dtr, current.x)
{
  if (!is.matrix(current.x) || ncol(current.x) < 2) {
    current.x <- cbind(current.x, 0)
  }

  action <- .Call("R_apply_rule", current.dtr, current.x)
  attr(current.dtr, "options")[action + 1L]
}


print.regime <- function(dtr, variables)
{
  for (i.stage in seq_along(dtr)) {
    cat(sprintf("Stage %d:\n", i.stage))
    d <- dtr[[i.stage]]
    op1 <- ifelse(substring(d[, "type"], 1, 1) == "L", "<=", ">")
    op2 <- ifelse(substring(d[, "type"], 2, 2) == "L", "<=", ">")
    var1 <- variables[d[, "j1"]]
    var2 <- variables[d[, "j2"]]
    term1 <- ifelse(is.finite(d[, "c1"]), paste(var1, op1, d[, "c1"]), "")
    term2 <- ifelse(is.finite(d[, "c2"]), paste(var2, op2, d[, "c2"]), "")
    cond <- ifelse(nchar(term1) > 0 & nchar(term2) > 0,
                   paste(term1, "and", term2), paste0(term1, term2))
    act <- attr(d, "options")[d[, "a"] + 1]
    cat(paste0("if ", cond, " then ", act, "\n"))
  }
  invisible(NULL)
}




dtr <- estimate.regime(y, a, x, stage.x, seed = 20)
print.regime(dtr, variables)
  
